"""
Defaults
--------
"""

from math import sqrt
from typing import Optional

import metacells.utilities.typing as utt

#: The generic minimal "significant" gene fraction. See
#: :py:func:`metacells.tools.high.find_high_fraction_genes`.
significant_gene_fraction: float = 1e-5

#: The generic minimal "significant" gene normalized variance. See
#: :py:func:`metacells.tools.high.find_high_normalized_variance_genes`.
significant_gene_normalized_variance: float = 2**2.5

#: The generic minimal "significant" gene relative variance. See
#: :py:func:`metacells.tools.high.find_high_relative_variance_genes`.
significant_gene_relative_variance: float = 0.1

#: The generic minimal "significant" gene similarity. See
#: :py:const:`bursty_lonely_max_gene_similarity`
#: and
#: :py:const:`rare_min_module_correlation`.
significant_gene_similarity: float = 0.1

#: The generic "significant" fold factor. See
#: :py:const:`deviants_min_gene_fold_factor`
#: :py:const:`rare_min_convincing_gene_fold_factor`,
#: and
#: :py:const:`dissolve_min_convincing_gene_fold_factor`.
significant_gene_fold_factor: float = 3.0

#: The generic additional "significant" fold factor for noisy genes, in addition to
#: :py:const:`significant_gene_fold_factor`. See
#: :py:const:`deviants_min_gene_fold_factor`
#: and
#: :py:const:`rare_min_convincing_gene_fold_factor`,
#: and
#: :py:const:`dissolve_min_convincing_gene_fold_factor`.
significant_noisy_gene_fold_factor: float = 2.0

#: The generic minimal value (number of UMIs) we can say is "significant" given the technical noise. See
#: :py:const:`rare_min_gene_maximum`,
#: :py:const:`rare_min_cell_module_total`,
#: and
#: :py:const:`cells_similarity_value_regularization`.
significant_value: int = 7

#: The generic minimal samples to use for downsampling the cells for some purpose. See
#: :py:const:`bursty_lonely_downsample_min_samples`,
#: :py:const:`select_downsample_min_samples`,
#: and
#: :py:func:`metacells.tools.downsample.downsample_cells`.
downsample_min_samples: int = 750

#: The generic minimal quantile of the cells total size to use for downsampling the cells for some
#: purpose. See
#: :py:const:`bursty_lonely_downsample_min_cell_quantile`,
#: :py:const:`select_downsample_min_cell_quantile`,
#: and
#: :py:func:`metacells.tools.downsample.downsample_cells`.
downsample_min_cell_quantile: float = 0.05

#: The generic maximal quantile of the cells total size to use for downsampling the cells for some
#: purpose. See
#: :py:const:`bursty_lonely_downsample_max_cell_quantile`,
#: :py:const:`select_downsample_max_cell_quantile`,
#: and
#: :py:func:`metacells.tools.downsample.downsample_cells`.
downsample_max_cell_quantile: float = 0.5

#: The window size to use to compute relative variance. See
#: :py:func:`metacells.utilities.computation.relative_variance_per`
#: and
#: :py:func:`metacells.tools.high.find_high_relative_variance_genes`.
relative_variance_window_size: int = 100

#: The method to use to compute similarities. See
#: :py:func:`metacells.tools.similarity.compute_obs_obs_similarity`,
#: and
#: :py:func:`metacells.tools.similarity.compute_var_var_similarity`.
similarity_method: str = "abs_pearson"

#: The default location for the logistics function. See
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`,
#: :py:func:`metacells.tools.similarity.compute_obs_obs_similarity`,
#: :py:func:`metacells.tools.similarity.compute_var_var_similarity`.
#: and
#: :py:func:`metacells.utilities.computation.logistics`.
logistics_location: float = 0.8

#: The default slope for the logistics function. See
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`,
#: :py:func:`metacells.tools.similarity.compute_obs_obs_similarity`,
#: :py:func:`metacells.tools.similarity.compute_var_var_similarity`.
#: and
#: :py:func:`metacells.utilities.computation.logistics`.
logistics_slope: float = 0.5

#: The minimal target number of observations (cells) in a pile, allowing us to
#: directly compute groups (metacells) for it. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
min_target_pile_size: int = 8000

#: The maximal target number of observations (cells) in a pile, allowing us to
#: directly compute groups (metacells) for it. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
max_target_pile_size: int = 24000

#: The target number of metacells computed in each pile. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
target_metacells_in_pile: int = 100

#: The generic target total metacell size (in cells). See
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
target_metacell_size: int = 48

#: The generic target total metacell total UMIs. See
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
target_metacell_umis: int = 160000

#: Whether to use geometrical mean to compute metacell gene fractions.
#: See :py:func:`metacells.pipeline.collect.collect_metacells`.
metacell_geo_mean: bool = True

#: The number of UMIs to use for regularization when computing metacell gene fractions geometrical mean.
#: See :py:func:`metacells.pipeline.collect.collect_metacells`.
metacell_umis_regularization: float = 1 / 16.0

#: The quantile of the cell sizes to use for computing the number of zero-valued cells.
#: See
#: :py:func:`metacells.pipeline.collect.collect_metacells`.
zeros_cell_size_quantile: float = 0.1

#: The number of UMIs in each cell for computing each metacell's total UMIs. See
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`
#: and
#: :py:func:`metacells.pipeline.collect.collect_metacells`.
cell_umis: Optional[utt.NumpyVector] = None

#: The generic maximal group size factor, above which we should split it. See
#: :py:const:`piles_min_split_size_factor`
#: and
#: :py:const:`candidates_min_split_size_factor`.
min_split_size_factor: float = 2.0

#: The generic minimal group size factor, below which we should consider dissolving it. See
#: :py:const:`piles_min_robust_size_factor`
#: and
#: :py:const:`dissolve_min_robust_size_factor`.
min_robust_size_factor: float = 0.5

#: The generic maximal group size factor, below which we should merge it. See
#: :py:const:`piles_max_merge_size_factor`,
#: :py:const:`candidates_max_merge_size_factor`.
max_merge_size_factor: float = 0.5

#: The minimal number of cells in a metacell, below which we would merge it. See
#: :py:const:`rare_min_cells_of_modules`,
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`
#: and
#: :py:func:`metacells.tools.dissolve.dissolve_metacells`.
min_metacell_size: int = 12

#: The maximal strength of a min-cut of a metacell that will cause it to be split. See
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`.
max_split_min_cut_strength: float = 0.1

#: The minimal number of cells to keep as a seed when cutting a metacell. See
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`.
min_cut_seed_cells: int = 7

#: The minimal size factor of a pile, above which we can split it. See
#: :py:const:`min_split_size_factor`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
piles_min_split_size_factor: float = 1.5

#: The minimal pile size factor, below which we should consider dissolving it. See
#: :py:const:`min_robust_size_factor`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
piles_min_robust_size_factor: float = 0.3

#: The maximal size factor of a pile, below which we should merge it. See
#: :py:const:`min_robust_size_factor`,
#: :py:const:`max_merge_size_factor`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
piles_max_merge_size_factor: float = 0.15

#: The minimal total value for a cell to be considered "properly sampled". See
#: :py:func:`metacells.tools.properly_sampled.find_properly_sampled_cells`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
#:
#: .. note::
#:
#:    There's no "reasonable" default value here. This must be tailored to the data.
properly_sampled_min_cell_total: Optional[int] = None

#: The maximal total value for a cell to be considered "properly sampled". See
#: :py:func:`metacells.tools.properly_sampled.find_properly_sampled_cells`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
#:
#: .. note::
#:
#:    There's no "reasonable" default value here. This must be tailored to the data.
properly_sampled_max_cell_total: Optional[int] = None

#: The minimal total value for a gene to be considered "properly sampled". See
#: :py:func:`metacells.tools.properly_sampled.find_properly_sampled_genes`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
properly_sampled_min_gene_total: int = 1

#: The maximal fraction of excluded gene UMIs from a cell for it to be considered
#: "properly_sampled". See
#: :py:func:`metacells.tools.properly_sampled.find_properly_sampled_cells`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
#:
#: .. note::
#:
#:    There's no "reasonable" default value here. This must be tailored to the data.
properly_sampled_max_excluded_genes_fraction: Optional[float] = None

#: The number of randomly selected cells to use for computing related genes. See
#: :py:func:`metacells.pipeline.related_genes.relate_to_lateral_genes`.
related_max_sampled_cells: int = 10000

#: How to compute gene-gene similarity for computing the related genes. See
#: :py:func:`metacells.pipeline.related_genes.relate_to_lateral_genes`.
related_genes_similarity_method: str = "abs_pearson"

#: The hierarchical clustering method to use for computing the related genes. See
#: :py:func:`metacells.pipeline.related_genes.relate_to_lateral_genes`.
related_genes_cluster_method: str = "ward"

#: The maximal number of genes in a related gene module. See
#: :py:func:`metacells.pipeline.related_genes.relate_to_lateral_genes`.
related_max_genes_of_modules: int = 64

#: The minimal mean fraction of a related gene. See
#: :py:func:`metacells.pipeline.related_genes.relate_to_lateral_genes`.
related_min_mean_gene_fraction: float = 1e-5

#: The minimal correlation of a related gene to the base gene(s). See
#: :py:func:`metacells.pipeline.related_genes.relate_to_lateral_genes`.
related_min_gene_correlation: float = 0.1

#: The number of randomly selected cells to use for computing "bursty lonely" genes. See
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
bursty_lonely_max_sampled_cells: int = 10000

#: The minimal samples to use for downsampling the cells for computing "bursty lonely" genes. See
#: :py:const:`downsample_min_cell_quantile`,
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes`,
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
bursty_lonely_downsample_min_samples: int = downsample_min_samples

#: The minimal quantile of the cells total size to use for downsampling the cells for computing
#: "bursty lonely" genes. See
#: :py:const:`downsample_min_cell_quantile`,
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes`,
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
bursty_lonely_downsample_min_cell_quantile: float = downsample_min_cell_quantile

#: The maximal quantile of the cells total size to use for downsampling the cells for computing
#: "bursty lonely" genes. See
#: :py:const:`downsample_min_cell_quantile`,
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes`,
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
bursty_lonely_downsample_max_cell_quantile: float = downsample_max_cell_quantile

#: The minimal total UMIs in the downsamples selected cells of a gene to be considered when
#: computing "lonely" genes. See
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes` and
#: :py:func:`metacells.tools.high.find_high_total_genes`.
bursty_lonely_min_gene_total: int = 100

#: The minimal normalized variance of a gene to be considered "bursty". See
#: :py:const:`significant_gene_normalized_variance`,
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
bursty_lonely_min_gene_normalized_variance: float = significant_gene_normalized_variance

#: The maximal similarity between a gene and another gene to be considered "lonely". See
#: :py:const:`significant_gene_similarity`,
#: :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes`
#: and
#: :py:func:`metacells.pipeline.exclude.extract_clean_data`.
bursty_lonely_max_gene_similarity: float = significant_gene_similarity

#: The maximal number of candidate rare genes. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_max_genes: int = 500

#: The maximal fraction of the cells where a gene is expressed to be considered "rare". See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_max_gene_cell_fraction: float = 1e-3

#: The minimal maximum-across-all-cells value of a gene to be considered as a candidate for rare
#: gene modules. See
#: :py:const:`significant_value`,
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_gene_maximum: int = significant_value

#: How to compute gene-gene similarity for computing the rare gene modules. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_genes_similarity_method: str = "repeated_pearson"

#: The hierarchical clustering method to use for computing the rare gene modules. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_genes_cluster_method: str = "ward"

#: The minimal number of genes in a rare gene module. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_genes_of_modules: int = 4

#: The minimal number of cells in a rare gene module. See
#: :py:const:`min_metacell_size`,
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_cells_of_modules: int = min_metacell_size

#: The maximal mean number of cells (as a fraction of the target metacell size) in a random pile for a rare gene module
#: to be considered rare. See
#: :py:const:`min_metacell_size`,
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_max_cells_factor_of_random_pile: float = 0.5

#: The minimal average correlation between the genes in a rare gene module. See
#: :py:func:`metacells.parameters.significant_gene_similarity`,
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_module_correlation: float = significant_gene_similarity

#: The minimal fold factor between rare cells and the rest of the population for a gene to be
#: considered related to the rare gene module. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_related_gene_fold_factor: float = 7

#: The maximal ratio of total cells to include as a result of adding a related gene to a rare gene
#: module. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_max_related_gene_increase_factor: float = 4.0

#: The minimal number of UMIs of a rare gene module in a cell to be considered as expressing the
#: rare behavior. See
#: :py:func:`metacells.tools.rare.find_rare_gene_modules`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_cell_module_total: int = int((significant_value + 1) / 2)

#: The maximal fraction of cells to mark as "deviants" in rare gene module piles. See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_deviants_max_cell_fraction: Optional[float] = 0.25

#: Whether to compute metacells more quickly with lower quality. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
quick_and_dirty: bool = False

#: The minimal samples to use for downsampling the cells for computing "select" genes. See
#: :py:const:`downsample_min_samples`,
#: :py:func:`metacells.tools.downsample.downsample_cells`,
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_downsample_min_samples: int = downsample_min_samples

#: The minimal quantile of the cells total size to use for downsampling the cells for computing
#: "select" genes. See
#: :py:const:`downsample_min_cell_quantile`,
#: :py:func:`metacells.tools.downsample.downsample_cells`,
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_downsample_min_cell_quantile: float = downsample_min_cell_quantile

#: The maximal quantile of the cells total size to use for downsampling the cells for computing
#: "select" genes. See
#: :py:const:`downsample_max_cell_quantile`,
#: :py:func:`metacells.tools.downsample.downsample_cells`,
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_downsample_max_cell_quantile: float = downsample_max_cell_quantile

#: The minimal relative variance of a gene to be considered a "select". See
#: :py:func:`metacells.tools.high.find_high_relative_variance_genes`,
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_min_gene_relative_variance: Optional[float] = 0.1

#: The minimal number of downsampled UMIs of a gene to be "select". See
#: :py:func:`metacells.tools.high.find_high_total_genes`,
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_min_gene_total: Optional[int] = 50

#: The minimal number of the top-3rd downsampled UMIs of a gene to be "select". See
#: :py:func:`metacells.tools.high.find_high_topN_genes`,
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_min_gene_top3: Optional[int] = 4

#: The minimal number of "select" genes. See
#: :py:func:`metacells.pipeline.select.extract_selected_data`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
select_min_genes: int = 100

#: Whether to compute cell-cell similarity using the log (base 2) of the data. See
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
cells_similarity_log_data: bool = True

#: The regularization factor to use if/when computing the fractions of the data for directly
#: computing the metacells. See
#: :py:const:`significant_value`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
cells_similarity_value_regularization: float = 1 / significant_value

#: The method to use to compute cell-cell similarity. See
#: :py:func:`metacells.tools.similarity.compute_obs_obs_similarity`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
cells_similarity_method: str = similarity_method

#: Whether to compute group-group similarity using the log (base 2) of the data. See
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
groups_similarity_log_data: bool = True

#: The method to use to compute group-group similarity. See
#: :py:const:`cells_similarity_method`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
groups_similarity_method: str = similarity_method

#: The target K for building the K-Nearest-Neighbors graph. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
knn_k: Optional[int] = None

#: The size of the default K for building the K-Nearest-Neighbors graph when computing metacells, multiplied
#: by the median number of cells needed to reach the :py:const:`target_metacell_size`. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`,
#: and
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`.
candidates_knn_k_size_factor: int = 2

#: The size of the default K for building the K-Nearest-Neighbors graph when computing groups of metacells, multiplied
#: by the median number of cells needed to reach the :py:const:`target_metacell_size`. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`,
#: and
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`.
piles_knn_k_size_factor: int = 3

#: The size of the default K for building the K-Nearest-Neighbors graph, based
#: on the number of cells needed to reach the :py:const:`target_metacell_umis` for a quantile of the cells. See
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`,
#: and
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`.
knn_k_umis_quantile: float = 0.1

#: The minimal target K for building the K-Nearest-Neighbors graph. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
min_knn_k: Optional[int] = 30

#: The factor of K edge ranks to keep when computing the balanced ranks for the
#: K-Nearest-Neighbors graph. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
knn_balanced_ranks_factor: float = sqrt(10)

#: The factor of K of edges to keep when pruning the incoming edges of the K-Nearest-Neighbors
#: graph. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
knn_incoming_degree_factor: float = 3.0

#: The factor of K of edges to keep when pruning the outgoing edges of the K-Nearest-Neighbors
#: graph. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
knn_outgoing_degree_factor: float = 1.0

#: The minimal outgoing degree of nodes in the KNN graph. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`
#: :py:func:`metacells.pipeline.umap.compute_knn_by_markers`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
knn_min_outgoing_degree: int = 2

#: The minimal outgoing degree of nodes in the markers KNN graph. See
#: :py:func:`metacells.pipeline.umap.compute_knn_by_markers` and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
markers_knn_min_outgoing_degree: int = 1

#: The minimal quantile of a seed to be selected. See
#: :py:func:`metacells.tools.candidates.choose_seeds`,
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
min_seed_size_quantile: float = 0.85

#: The maximal quantile of a seed to be selected. See
#: :py:func:`metacells.tools.candidates.choose_seeds`,
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
max_seed_size_quantile: float = 0.95

#: By how much (as a fraction) to cooldown the temperature after doing a pass on all the nodes. See
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells` and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
cooldown_pass: float = 0.02

#: By how much (as a fraction) to cooldown the node temperature after improving it. See
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells` and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
cooldown_node: float = 0.25

#: By how much (as a fraction) to reduce the cooldown each time we re-optimize a slightly modified
#: partition. See
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells` and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
cooldown_phase: float = 0.75

#: The minimal size factor of clusters to split when clustering the nodes of the
#: K-Nearest-Neighbors graph. See
#: :py:const:`min_split_size_factor`,
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
candidates_min_split_size_factor: float = min_split_size_factor

#: The maximal size factor of clusters to merge when clustering the nodes of the
#: K-Nearest-Neighbors graph. See
#: :py:const:`max_merge_size_factor`,
#: :py:func:`metacells.tools.candidates.compute_candidate_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
candidates_max_merge_size_factor: float = max_merge_size_factor

#: The policy to use for deciding which cell is "deviant". See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_policy: str = "gaps"

#: If using the ``gaps`` deviants policy, how many cells to skip ahead when computing gap sizes
#: (``0``, ``1``, ``2``). See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_gap_skip_cells: int = 1

#: Do not mark deviants by a gene in a metacell if it causes more than this number of cells to become deviant
#: (unless the count is no more than :py:const:`deviants_max_gap_cells_fraction`). See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_max_gap_cells_count: int = 3

#: Do not mark deviants by a gene in a metacell if it causes more than this fraction of cells to become deviant
#: (unless the count is no more than :py:const:`deviants_max_gap_cells_count`). See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_max_gap_cells_fraction: float = 0.1

#: Use this quantile to pick the minimal cell size for the regularization factor when
# computing deviants. See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviant_cells_regularization_quantile: float = 0.25

#: The minimal fold factor for a gene to indicate a cell is "deviant". See
#: :py:const:`significant_gene_fold_factor`,
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_min_gene_fold_factor: float = significant_gene_fold_factor

#: The minimal number of UMIs in the gene for two cells for using it as a certificate for a gap between
#: the expression level of the gene in the cells for computinng deviants. See
#: :py:const:`significant_gene_fold_factor`,
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_min_compare_umis: int = 8

#: The minimal additional fold factor for a noisy gene to indicate a cell is "deviant",
# in addition to :py:const:`deviants_min_gene_fold_factor`. See
#: :py:const:`significant_gene_fold_factor`,
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_min_noisy_gene_fold_factor: float = significant_noisy_gene_fold_factor

#: The maximal fraction of genes to use to indicate cell are "deviants". See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_max_gene_fraction: float = 0.03

#: The maximal fraction of cells to mark as "deviants". See
#: :py:func:`metacells.tools.deviants.find_deviant_cells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
deviants_max_cell_fraction: Optional[float] = 0.25

#: The minimal size factor for a metacell to be considered "robust". See
#: :py:const:`min_robust_size_factor`
#: :py:func:`metacells.tools.dissolve.dissolve_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
dissolve_min_robust_size_factor: float = min_robust_size_factor

#: The minimal size factor for a metacell to be considered "robust" when grouping rare gene module
#: cells. See
#: :py:const:`min_robust_size_factor`,
#: :py:const:`max_merge_size_factor`
#: :py:func:`metacells.tools.dissolve.dissolve_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_dissolve_min_robust_size_factor: float = max_merge_size_factor

#: The minimal fold factor of a gene in a rare metacell to make it "convincing". See
#: module cells. See
#: :py:const:`max_merge_size_factor`
#: :py:func:`metacells.tools.dissolve.dissolve_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
rare_min_convincing_gene_fold_factor: Optional[float] = None

#: The minimal fold factor of a gene in a metacell to make it "convincing". See
#: :py:const:`significant_gene_fold_factor`,
#: :py:func:`metacells.tools.dissolve.dissolve_metacells`,
#: :py:func:`metacells.pipeline.direct.compute_direct_metacells`,
#: :py:func:`metacells.pipeline.divide_and_conquer.compute_divide_and_conquer_metacells`
#: and
#: :py:func:`metacells.pipeline.divide_and_conquer.divide_and_conquer_pipeline`.
dissolve_min_convincing_gene_fold_factor: float = significant_gene_fold_factor

#: The number of most-distinct genes to collect for each cell. See
#: :py:func:`metacells.tools.distinct.find_distinct_genes`.
distinct_genes_count: int = 20

#: The maximal number of marker genes to use for UMAP.
#: See
#: :py:func:`metacells.pipeline.umap.compute_knn_by_markers` and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_max_marker_genes: int = 1000

#: Whether to ignore lateral genes when computing the UMAP.
#: See
#: :py:func:`metacells.pipeline.umap.compute_knn_by_markers` and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_ignore_lateral_genes: bool = True

#: The regularization factor to use if/when computing the fractions of the data for UMAP.
#: See
#: :py:const:`metacells.parameters.significant_gene_fraction`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_similarity_value_regularization: float = significant_gene_fraction

#: Whether to compute metacell-metacell similarity using the log (base 2) of the data for UMAP. See
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_similarity_log_data: bool = True

#: The method to use to compute similarities for UMAP. See
#: :py:func:`metacells.tools.similarity.compute_obs_obs_similarity`,
#: :py:func:`metacells.tools.similarity.compute_var_var_similarity`,
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_similarity_method: str = "logistics_abs_pearson"

#: The minimal UMAP point distance. See :py:const:`umap_spread` and
#: :py:func:`metacells.tools.layout.umap_by_distances`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_min_dist: float = 0.5

#: The minimal UMAP spread. This is automatically raised if the :py:const:`umap_min_dist` is higher.
#: See :py:func:`metacells.tools.layout.umap_by_distances`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_spread: float = 1.0

#: The UMAP KNN graph degree. See
#: :py:func:`metacells.tools.layout.umap_by_distances`
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_k: int = 15

#: The UMAP KNN skeleton graph degree. See
#: :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph`,
#: :py:func:`metacells.tools.knn_graph.compute_var_var_knn_graph`,
#: and
#: :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
skeleton_k: int = 4

#: The value to add to gene fractions before applying the log function. See
#: See :py:func:`metacells.pipeline.umap.compute_umap_by_markers`.
umap_fraction_regularization: float = 1e-5

#: The fraction of the UMAP plot area to cover with points. See
#: :py:func:`metacells.utilities.computation.cover_diameter`,
#: :py:func:`metacells.utilities.computation.cover_coordinates`
#: and
#: :py:func:`metacells.tools.layout.umap_by_distances`,
spread_cover_fraction: float = 1 / 3.0

#: The noise to add to the UMAP plot area. See
#: :py:func:`metacells.utilities.computation.cover_coordinates`
#: and
#: :py:func:`metacells.tools.layout.umap_by_distances`,
spread_noise_fraction: float = 0.1

#: The minimal total number of UMIs for a gene to compute meaningful quality statistics for it.
#: See :py:func:`metacells.tools.quality.compute_stdev_logs`,
#: :py:func:`metacells.tools.quality.compute_inner_folds`, and
#: :py:func:`metacells.tools.quality.compute_outliers_fold_factors`.
quality_min_gene_total: int = 40

#: The maximal amount of memory to use when guessing the number of parallel piles. If zero or
#: negative, is the fraction of the machine's total RAM to use as a safety buffer. See
#: :py:func:`metacells.pipeline.divide_and_conquer.guess_max_parallel_piles`.
max_gbs: float = -0.1

#: Whether the projection will ignore genes where the ranges of the corrected vs. the projected expression is too low.
#: See :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_filter_ranges: bool = True

#: The quantile to use (on both low and high ends) to compute the range of expression of a corrected of a projected
#: gene. See :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_ignore_range_quantile: float = 0.02

#: The minimal overlap (shared range divided by query range) for genes to keep projecting. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_ignore_range_min_overlap_fraction: float = 0.5

#: The minimal fraction of the query marker genes that are fitted for a query metacell to be considered "similar" to the
#: atlas. See :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_min_query_markers_fraction: float = 1 / 3

#: The regularization factor to use when computing fold factors for projecting a query onto an atlas. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_fold_regularization: float = 1e-5

#: The minimal number of UMIs for a gene to be a potential cause to mark a metacell as dissimilar. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_min_significant_gene_umis: int = 40

#: The number of atlas candidates to consider when projecting a query onto an atlas. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_candidates_count: int = 50

#: The minimal number of atlas candidates to use even if they fail the consistency check as a fraction of
#: :py:const:`project_candidates_count`. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_min_candidates_fraction: float = 1.0 / 3.0

#: The minimal weight of an atlas metacell used for the projection of a query metacell. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_min_usage_weight: float = 1e-5

#: The maximal fold factor of genes between the projection and the query metacell. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_max_projection_fold_factor: float = significant_gene_fold_factor

#: The maximal additional fold factor of noisy genes between the projection and the query metacell,
# in addition to :py:const:`project_max_projection_fold_factor`. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_max_projection_noisy_fold_factor: float = significant_noisy_gene_fold_factor

#: The maximal fold factor of genes between the atlas metacells used for the projection of a query metacell. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_max_consistency_fold_factor: float = significant_gene_fold_factor - 1.0

#: Whether to compute projectio the log (base 2) of the data. See
#: :py:func:`metacells.tools.project.compute_projection_weights`.
project_log_data: bool = True

#: The regularization factor to use when computing log of fractions for finding the most similar group for outliers. See
#: :py:func:`metacells.tools.quality.compute_outliers_matches`.
outliers_fold_regularization: float = 1e-5

#: Whether to ignore the lateral genes of the atlas when computing projections. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
ignore_atlas_lateral_genes: bool = True

#: Whether to ignore the noisy genes of the atlas when computing projections. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
consider_atlas_noisy_genes: bool = True

#: Whether to ignore the non-marker genes of the atlas when computing projections. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
only_atlas_marker_genes: bool = True

#: Whether to ignore the non-marker genes of the query when computing projections. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
only_query_marker_genes: bool = False

#: Whether to ignore the lateral genes of the query when computing projections. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
ignore_query_lateral_genes: bool = True

#: Whether to ignore the noisy genes of the query when computing projections. See
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
consider_query_noisy_genes: bool = True

#: The minimal fraction of metacells where a gene has a high projection fold factor to mark the gene as "misfit".
#: See :py:func:`metacells.pipeline.projection.projection_pipeline`.
misfit_min_metacells_fraction: float = 0.5

#: The minimal fold between the maximal and minimal gene expression in metacells to be a "marker".
#: See :py:func:`metacells.tools.high.find_metacells_marker_genes`.
min_marker_metacells_gene_range_fold_factor: float = 2.0

#: The regularization factor to use after computing the fractions of the data for
#: computing metacell gene range folds. See
#: :py:func:`metacells.tools.high.find_metacells_marker_genes`.
metacells_gene_range_regularization: float = 1e-5

#: The minimal maximal gene expression in metacells to be a "marker".
#: See :py:func:`metacells.tools.high.find_metacells_marker_genes`.
min_marker_max_metacells_gene_fraction: float = 1e-4

#: Whether to renormalize the query to account for missing atlas genes when computing projections.
#: :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_renormalize_query: bool = False

#: Whether to compute linear corrections for genes between the query and the atlas.
#: See :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_corrections: bool = False

#: The minimal correlation between observed and projected genes for considering linear correction of the query gene
#: value. See :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_min_corrected_gene_correlation: float = 0.8

#: The minimal strength of the correction between the mean query and projected mean value (for correlated genes).
#: See :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_min_corrected_gene_factor: float = 0.15

#: The maximal number of deviant genes allowed for saying a query is similar to the projection in the atlas.
#: See :py:func:`metacells.tools.quality.compute_similar_query_metacells`
# and :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_max_misfit_genes: int = 3

#: The quantile of each gene's normalized variance across the metacells to use for the overall gene's variability.
#: See :py:func:`metacells.tools.quality.compute_type_genes_normalized_variances`.
type_gene_normalized_variance_quantile: float = 0.95

#: Minimal fraction of atlas essential genes required for saying a query is similar to the projection in the
#: atlas.
#: See :py:func:`metacells.tools.quality.compute_similar_query_metacells`
# and :py:func:`metacells.pipeline.projection.projection_pipeline`.
project_min_essential_genes_fraction: float = 0.75



"""
Functions for analysis tools.

Tools take as input some annotated data and either return some computed results, or write the
results as new annotations within the same data, or return a new annotated data object containing
the results. Tools are meant to be composable into a complete processing
:py:mod:`metacells.pipeline`, typically by having one tool create annotation with "agreen upon"
name(s) and another further processing them. While this is flexible and convenient, there is no
static typing or analysis to ensure that the input of the next tool was actually created by previous
tool(s).

All the functions included here are exported under ``metacells.tl``.
"""

from .apply import *
from .bursty_lonely import *
from .candidates import *
from .convey import *
from .deviants import *
from .dissolve import *
from .distinct import *
from .downsample import *
from .filter import *
from .group import *
from .high import *
from .knn_graph import *
from .layout import *
from .mask import *
from .named import *
from .project import *
from .properly_sampled import *
from .quality import *
from .rare import *
from .similarity import *



"""
Deviants
--------
"""

import sys
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore
from scipy import sparse as sp  # type: ignore
from scipy import stats as st

import metacells.parameters as pr
import metacells.utilities as ut

if "sphinx" not in sys.argv[0]:
    import metacells.extensions as xt  # type: ignore

__all__ = [
    "find_deviant_cells",
]


@ut.logged(candidates=ut.groups_description)
@ut.timed_call()
@ut.expand_doc()
def find_deviant_cells(  # pylint: disable=too-many-statements,too-many-branches
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    candidates: Union[str, ut.Vector] = "candidate",
    min_gene_fold_factor: float = pr.deviants_min_gene_fold_factor,
    min_compare_umis: int = pr.deviants_min_compare_umis,
    gap_skip_cells: int = pr.deviants_gap_skip_cells,
    min_noisy_gene_fold_factor: float = pr.deviants_min_noisy_gene_fold_factor,
    max_gene_fraction: float = pr.deviants_max_gene_fraction,
    max_cell_fraction: Optional[float] = pr.deviants_max_cell_fraction,
    max_gap_cells_count: int = pr.deviants_max_gap_cells_count,
    max_gap_cells_fraction: float = pr.deviants_max_gap_cells_fraction,
    cells_regularization_quantile: float = pr.deviant_cells_regularization_quantile,
    policy: str = pr.deviants_policy,
) -> ut.Vector:
    """
    Find cells which are have significantly different gene expression from the metacells they are
    belong to based on ``what`` (default: {what}) data.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    Obeys (ignores the genes of) the ``noisy_gene`` per-gene (variable) annotation, if any.

    The exact method depends on the ``policy`` (one of ``gaps`` or ``max``). By default we use the ``gaps`` policy as it
    gives a much lower fraction of deviants at a minor cost in the variance inside each metacell. The ``max`` policy
    provides the inverse trade-off, giving slightly more consistent metacells at the cost of a much higher fraction of
    deviants.

    **Returns**

    A boolean mask of all the cells which should be considered "deviant".

    **Gaps Computation Parameters**

    Intuitively, for each gene for each metacell we can look at the sorted expression level of the gene in all the
    metacell's cells. We look for a large gap between a few low-expressing or high-expressing cells and the rest of the
    cells. If we find such a gap, the few cells below or above it are considered to be deviants.

    1. For each gene in each cell of each metacell, compute the log (base 2) of the fraction of the gene's
       UMIs out of the total UMIs of the metacell, with a 1-UMI regularization factor.

    2. Sort the expression level of each gene in each metacell.

    3. Look for a gap of at least ``min_gene_fold_factor`` (default: {min_gene_fold_factor}), or for ``noisy_gene``,
       an additional ``min_noisy_gene_fold_factor`` (default: {min_noisy_gene_fold_factor}) between the sorted gene
       expressions. If ``gap_skip_cells`` (default: {gap_skip_cells}) is 0, look for a gap between consecutive sorted
       cell expression levels. If it is 1 or 2, skip this number of entries. Ignore gaps if the total number of UMIs
       of the gene in the two compared cells is less than ``min_compare_umis`` (default: {min_compare_umis}).

    4. Ignore gaps that cause more than ``max_gap_cells_fraction`` (default: {max_gap_cells_fraction}) and also
       more than ``max_gap_cells_count`` (default: {max_gap_cells_count}) to be separated. That is, a single gene
       can only mark as deviants  "a few" cells of the metacell.

    5. If any cells were marked as deviants, re-run the above, ignoring any cells previously marked as deviants.

    6. If the total number of cells is more than ``max_cell_fraction`` (default: {max_cell_fraction}) of the cells,
       increase ``min_gene_fold_factor`` by 0.15 (~x1.1) and try again from the top.

    **Max Computation Parameters**

    1. Compute for each candidate metacell the median fraction of the UMIs expressed by each gene.
       Scale this by each cell's total UMIs to compute the expected number of UMIs for each cell.
       Compute the fold factor log2((actual UMIs + 1) / (expected UMIs + 1)) for each gene for each
       cell.

    2. Compute the excess fold factor for each gene in each cell by subtracting ``min_gene_fold_factor`` (default:
       {min_gene_fold_factor}) from the above. For ``noisy_gene``, also subtract ``min_noisy_gene_fold_factor`` to the
       threshold.

    4. For each cell, consider the maximal gene excess fold factor. Consider all
       cells with a positive maximal threshold as deviants. If more than ``max_cell_fraction`` (default:
       {max_cell_fraction}) of the cells have a positive maximal excess fold factor, increase the threshold from 0 so
       that only this fraction are marked as deviants.
    """
    assert 0 <= gap_skip_cells <= 2
    gap_skip_cells += 1
    # assert policy in ("max", "gaps")
    assert policy in ("votes", "count", "max", "sum", "gaps")
    if policy == "votes":
        return _votes_deviant_cells(
            adata,
            what,
            candidates=candidates,
            min_gene_fold_factor=min_gene_fold_factor,
            max_gene_fraction=max_gene_fraction or 1.0,
            max_cell_fraction=max_cell_fraction or 1.0,
        )

    if policy == "gaps":
        return _gaps_deviant_cells(
            adata,
            what,
            candidates=candidates,
            min_gene_fold_factor=min_gene_fold_factor,
            gap_skip_cells=gap_skip_cells,
            min_noisy_gene_fold_factor=min_noisy_gene_fold_factor,
            max_gap_cells_count=max_gap_cells_count,
            max_gap_cells_fraction=max_gap_cells_fraction,
            max_cell_fraction=max_cell_fraction or 1.0,
            cells_regularization_quantile=cells_regularization_quantile,
            min_compare_umis=min_compare_umis,
        )

    cells_count, genes_count = adata.shape
    assert cells_count > 0

    candidate_of_cells = ut.get_o_numpy(adata, candidates, formatter=ut.groups_description).copy()
    candidates_count = np.max(candidate_of_cells) + 1

    total_umis_per_cell = ut.get_o_numpy(adata, what, sum=True)
    assert total_umis_per_cell.size == cells_count

    noisy_genes_mask: Optional[ut.NumpyVector] = None
    if ut.has_data(adata, "noisy_gene"):
        noisy_genes_mask = ut.get_v_numpy(adata, "noisy_gene")
        if not np.any(noisy_genes_mask):
            noisy_genes_mask = None

    umis_per_gene_per_cell = ut.get_vo_proper(adata, what, layout="row_major")

    fold_per_gene_per_cell = np.zeros(adata.shape, dtype="float32")
    for candidate_index in range(candidates_count):
        candidate_cell_indices = np.where(candidate_of_cells == candidate_index)[0]

        candidate_cells_count = candidate_cell_indices.size
        assert candidate_cells_count > 0

        total_umis_per_grouped = total_umis_per_cell[candidate_cell_indices]
        umis_per_gene_per_grouped = ut.to_numpy_matrix(umis_per_gene_per_cell[candidate_cell_indices, :], copy=True)
        assert ut.is_layout(umis_per_gene_per_grouped, "row_major")
        assert umis_per_gene_per_grouped.shape == (candidate_cells_count, genes_count)

        fraction_per_gene_per_grouped = ut.fraction_by(umis_per_gene_per_grouped, by="row", sums=total_umis_per_grouped)

        fraction_per_gene_per_grouped_by_columns = ut.to_layout(fraction_per_gene_per_grouped, "column_major")
        assert fraction_per_gene_per_grouped_by_columns.shape == (candidate_cells_count, genes_count)
        median_fraction_per_gene = ut.median_per(fraction_per_gene_per_grouped_by_columns, per="column")
        assert len(median_fraction_per_gene) == genes_count

        _, dense, compressed = ut.to_proper_matrices(fraction_per_gene_per_grouped)

        if compressed is not None:
            if compressed.nnz == 0:
                continue

            extension_name = "fold_factor_compressed_%s_t_%s_t_%s_t" % (  # pylint: disable=consider-using-f-string
                compressed.data.dtype,
                compressed.indices.dtype,
                compressed.indptr.dtype,
            )
            extension = getattr(xt, extension_name)

            with ut.timed_step("extensions.fold_factor_compressed"):
                extension(
                    compressed.data,
                    compressed.indices,
                    compressed.indptr,
                    total_umis_per_grouped,
                    median_fraction_per_gene,
                )

            fold_per_gene_per_cell[candidate_cell_indices, :] = compressed

        else:
            assert dense is not None

            extension_name = f"fold_factor_dense_{dense.dtype}_t"
            extension = getattr(xt, extension_name)

            with ut.timed_step("extensions.fold_factor_dense"):
                extension(
                    dense,
                    total_umis_per_grouped,
                    median_fraction_per_gene,
                )

            fold_per_gene_per_cell[candidate_cell_indices, :] = dense

    effective_fold_per_gene_per_cell = np.abs(fold_per_gene_per_cell)
    effective_fold_per_gene_per_cell -= min_gene_fold_factor
    if noisy_genes_mask is not None:
        effective_fold_per_gene_per_cell[:, noisy_genes_mask] -= min_noisy_gene_fold_factor
    effective_fold_per_gene_per_cell[effective_fold_per_gene_per_cell < 0] = 0

    if policy == "count":
        policy_per_cell = ut.sum_per(effective_fold_per_gene_per_cell > 0, per="row")
    elif policy == "sum":
        policy_per_cell = ut.sum_per(effective_fold_per_gene_per_cell, per="row")
    elif policy == "max":
        policy_per_cell = ut.max_per(effective_fold_per_gene_per_cell, per="row")
    else:
        assert False
    ut.log_calc(f"{policy}_per_gene_per_cell", policy_per_cell)

    if max_cell_fraction is None or max_cell_fraction == 1.0:
        policy_threshold = 0
    else:
        policy_threshold = np.quantile(policy_per_cell, 1.0 - max_cell_fraction)
    ut.log_calc(f"{policy}_threshold", policy_threshold)

    deviant_cells_mask = policy_per_cell > policy_threshold
    ut.log_calc("deviant_cells_mask", deviant_cells_mask)
    return deviant_cells_mask


def _gaps_deviant_cells(  # pylint: disable=too-many-statements
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    candidates: Union[str, ut.Vector],
    min_gene_fold_factor: float,
    gap_skip_cells: int,
    min_noisy_gene_fold_factor: float,
    max_cell_fraction: float,
    max_gap_cells_count: int,
    max_gap_cells_fraction: float,
    cells_regularization_quantile: float,
    min_compare_umis: int,
) -> ut.Vector:
    assert min_gene_fold_factor > 0
    assert 0 < max_cell_fraction < 1
    cells_count = adata.n_obs
    genes_count = adata.n_vars

    min_gap_per_gene = np.full(genes_count, min_gene_fold_factor, dtype="float32")
    if ut.has_data(adata, "noisy_gene"):
        noisy_genes_mask = ut.get_v_numpy(adata, "noisy_gene")
        min_gap_per_gene[noisy_genes_mask] += min_noisy_gene_fold_factor

    candidate_per_cell = ut.get_o_numpy(adata, candidates, formatter=ut.groups_description).astype("int32")
    candidates_count = np.max(candidate_per_cell) + 1

    umis_per_gene_per_cell = ut.to_numpy_matrix(ut.get_vo_proper(adata, what, layout="row_major"), copy=True).astype(
        "float32"
    )

    regularization_per_cell = np.empty(cells_count, dtype="float32")
    max_gap_per_cell = np.empty(cells_count, dtype="float32")
    new_deviant_per_cell = np.empty(cells_count, dtype="bool")
    deviant_per_cell = np.empty(cells_count, dtype="bool")

    total_umis_per_cell = ut.get_o_numpy(adata, what, sum=True).copy()
    fraction_per_gene_per_cell = umis_per_gene_per_cell / total_umis_per_cell[:, np.newaxis]

    for candidate_index in range(candidates_count):
        mask_per_cell = candidate_per_cell == candidate_index
        total_umis_per_cell_of_candidate = total_umis_per_cell[mask_per_cell]
        regularization_of_candidate = max(
            1.0, np.quantile(total_umis_per_cell_of_candidate, cells_regularization_quantile)
        )
        regularization_per_cell_of_candidate = np.maximum(total_umis_per_cell_of_candidate, regularization_of_candidate)
        regularization_per_cell[mask_per_cell] = 1.0 / regularization_per_cell_of_candidate
    ut.log_calc("regularization_per_cell", regularization_per_cell, formatter=ut.sizes_description)

    log_fraction_per_gene_per_cell = np.log2(fraction_per_gene_per_cell + regularization_per_cell[:, np.newaxis])

    outer_repeat = 0
    while True:
        outer_repeat += 1
        ut.log_calc(f"min_gene_fold_factor ({outer_repeat}.*)", min_gene_fold_factor)

        new_deviant_per_cell[:] = True
        deviant_per_cell[:] = False
        deviants_fraction = 0.0

        inner_repeat = 0
        while deviants_fraction <= max_cell_fraction:
            inner_repeat += 1
            max_gap_per_cell[:] = 0.0
            with ut.timed_step("extensions.compute_cell_gaps"):
                extension = getattr(xt, "compute_cell_gaps")
                extension(
                    umis_per_gene_per_cell,
                    fraction_per_gene_per_cell,
                    log_fraction_per_gene_per_cell,
                    candidate_per_cell,
                    deviant_per_cell,
                    new_deviant_per_cell,
                    min_gap_per_gene,
                    candidates_count,
                    gap_skip_cells,
                    max_gap_cells_count,
                    max_gap_cells_fraction,
                    float(min_compare_umis),
                    max_gap_per_cell,
                )
            ut.log_calc(
                f"max_gap_per_cell ({outer_repeat}.{inner_repeat})", max_gap_per_cell, formatter=ut.sizes_description
            )

            large_gap_count = np.sum(max_gap_per_cell > 0.0)
            ut.log_calc(
                f"large_gap_count ({outer_repeat}.{inner_repeat})",
                large_gap_count,
                formatter=lambda count: ut.ratio_description(adata.n_obs, "cell", count, "large-gap"),
            )
            remaining_cell_fraction = 1 - max(max_cell_fraction - deviants_fraction, 0.0)
            if large_gap_count / adata.n_obs <= remaining_cell_fraction:
                gap_threshold = 0.0
            else:
                gap_threshold = np.quantile(max_gap_per_cell, remaining_cell_fraction)
            ut.log_calc(f"gap_threshold ({outer_repeat}.{inner_repeat})", gap_threshold)

            new_deviant_per_cell = (max_gap_per_cell > gap_threshold) & ~deviant_per_cell
            ut.log_calc(f"new_deviant_per_cell ({outer_repeat}.{inner_repeat})", new_deviant_per_cell)
            new_deviants_count = np.sum(new_deviant_per_cell)
            if new_deviants_count == 0:
                break
            deviants_fraction += new_deviants_count / adata.n_obs
            deviant_per_cell = deviant_per_cell | new_deviant_per_cell
            ut.log_calc(f"deviants_fraction ({outer_repeat}.{inner_repeat})", deviants_fraction)
            ut.log_calc(f"deviant_per_cell ({outer_repeat}.{inner_repeat})", deviant_per_cell)

        if deviants_fraction <= max_cell_fraction:
            ut.log_calc(f"final min_gene_fold_factor ({outer_repeat}.*)", min_gene_fold_factor)
            return deviant_per_cell

        min_gap_per_gene += 1 / 8
        min_gene_fold_factor += 1 / 8


def _votes_deviant_cells(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    candidates: Union[str, ut.Vector],
    min_gene_fold_factor: float,
    max_gene_fraction: float,
    max_cell_fraction: float,
) -> ut.Vector:
    assert min_gene_fold_factor > 0
    assert 0 < max_gene_fraction < 1
    assert 0 < max_cell_fraction < 1

    cells_count, genes_count = adata.shape
    assert cells_count > 0

    candidate_of_cells = ut.get_o_numpy(adata, candidates, formatter=ut.groups_description).copy()

    totals_of_cells = ut.get_o_numpy(adata, what, sum=True)
    assert totals_of_cells.size == cells_count

    noisy_genes_mask: Optional[ut.NumpyVector] = None
    if ut.has_data(adata, "noisy_gene"):
        noisy_genes_mask = ut.get_v_numpy(adata, "noisy_gene")

    data = ut.get_vo_proper(adata, what, layout="row_major")

    full_indices_of_remaining_cells = np.arange(cells_count)
    full_votes_of_deviant_cells = np.zeros(cells_count, dtype="int32")
    votes_of_deviant_genes = np.zeros(genes_count, dtype="int32")
    max_deviant_cells_count = cells_count * max_cell_fraction

    while True:
        ut.log_calc("max_deviant_cells_count", max_deviant_cells_count)
        acceptable_cells_mask = np.zeros(cells_count, dtype="bool")
        list_of_fold_factors, list_of_cell_index_of_rows = _collect_fold_factors(
            data=data,
            candidate_of_cells=candidate_of_cells,
            totals_of_cells=totals_of_cells,
            min_gene_fold_factor=min_gene_fold_factor,
            acceptable_cells_mask=acceptable_cells_mask,
            noisy_genes_mask=noisy_genes_mask,
        )

        fold_factors = _construct_fold_factors(cells_count, list_of_fold_factors, list_of_cell_index_of_rows)
        ut.log_calc("fold_factors_nnz", fold_factors.nnz)
        if fold_factors.nnz == 0:
            break

        deviant_gene_indices = _filter_genes(
            cells_count=cells_count,
            genes_count=genes_count,
            fold_factors=fold_factors,
            min_gene_fold_factor=min_gene_fold_factor,
            max_gene_fraction=max_gene_fraction,
        )

        deviant_genes_fold_ranks = _fold_ranks(
            cells_count=cells_count, fold_factors=fold_factors, deviant_gene_indices=deviant_gene_indices
        )

        votes_of_deviant_cells, did_reach_max_deviant_cells = _filter_cells(
            cells_count=cells_count,
            deviant_genes_fold_ranks=deviant_genes_fold_ranks,
            deviant_gene_indices=deviant_gene_indices,
            votes_of_deviant_genes=votes_of_deviant_genes,
            max_deviant_cells_count=max_deviant_cells_count,
        )

        full_votes_of_deviant_cells[full_indices_of_remaining_cells] = votes_of_deviant_cells
        remaining_cells_mask = votes_of_deviant_cells == 0

        ut.log_calc("did_reach_max_deviant_cells", did_reach_max_deviant_cells)
        ut.log_calc("surviving_cells_mask", remaining_cells_mask)

        if did_reach_max_deviant_cells or np.all(remaining_cells_mask):
            break

        max_deviant_cells_count -= np.sum(~remaining_cells_mask)
        assert max_deviant_cells_count > 0

        ut.log_calc("acceptable_cells_mask", acceptable_cells_mask)
        remaining_cells_mask &= ~acceptable_cells_mask
        ut.log_calc("remaining_cells_mask", remaining_cells_mask)
        if not np.any(remaining_cells_mask):
            break

        data = data[remaining_cells_mask, :]
        candidate_of_cells = candidate_of_cells[remaining_cells_mask]
        totals_of_cells = totals_of_cells[remaining_cells_mask]
        full_indices_of_remaining_cells = full_indices_of_remaining_cells[remaining_cells_mask]
        cells_count = len(full_indices_of_remaining_cells)

    deviant_cells_mask = full_votes_of_deviant_cells > 0
    ut.log_return("deviant_cells", deviant_cells_mask)
    return deviant_cells_mask


@ut.timed_call()
def _collect_fold_factors(  # pylint: disable=too-many-statements
    *,
    data: ut.ProperMatrix,
    candidate_of_cells: ut.NumpyVector,
    totals_of_cells: ut.NumpyVector,
    min_gene_fold_factor: float,
    acceptable_cells_mask: ut.NumpyVector,
    noisy_genes_mask: Optional[ut.NumpyVector],
) -> Tuple[List[ut.CompressedMatrix], List[ut.NumpyVector]]:
    list_of_fold_factors: List[ut.CompressedMatrix] = []
    list_of_cell_index_of_rows: List[ut.NumpyVector] = []

    cells_count, genes_count = data.shape
    candidates_count = np.max(candidate_of_cells) + 1

    ut.timed_parameters(candidates=candidates_count, cells=cells_count, genes=genes_count)
    remaining_cells_count = cells_count

    for candidate_index in range(candidates_count):
        candidate_cell_indices = np.where(candidate_of_cells == candidate_index)[0]

        candidate_cells_count = candidate_cell_indices.size
        if candidate_cells_count == 0:
            continue

        list_of_cell_index_of_rows.append(candidate_cell_indices)
        remaining_cells_count -= candidate_cells_count

        if candidate_cells_count < 2:
            compressed = sp.csr_matrix(
                ([], [], [0] * (candidate_cells_count + 1)), shape=(candidate_cells_count, genes_count)
            )
            list_of_fold_factors.append(compressed)
            assert compressed.has_sorted_indices
            assert compressed.has_canonical_format
            continue

        umis_of_candidate_by_row: ut.ProperMatrix = data[candidate_cell_indices, :]
        assert ut.is_layout(umis_of_candidate_by_row, "row_major")
        assert umis_of_candidate_by_row.shape == (candidate_cells_count, genes_count)

        totals_of_candidate_cells = totals_of_cells[candidate_cell_indices]

        fractions_of_candidate_by_row = ut.scale_by(
            umis_of_candidate_by_row, np.reciprocal(totals_of_candidate_cells), by="row"
        )

        fractions_of_candidate_by_column = ut.to_layout(fractions_of_candidate_by_row, "column_major")
        medians_of_candidate_genes = ut.median_per(fractions_of_candidate_by_column, per="column")

        _, dense, compressed = ut.to_proper_matrices(umis_of_candidate_by_row)

        if compressed is not None:
            if compressed.nnz == 0:
                list_of_fold_factors.append(compressed)
                continue

            extension_name = "fold_factor_compressed_%s_t_%s_t_%s_t" % (  # pylint: disable=consider-using-f-string
                compressed.data.dtype,
                compressed.indices.dtype,
                compressed.indptr.dtype,
            )
            extension = getattr(xt, extension_name)

            with ut.timed_step("extensions.fold_factor_compressed"):
                extension(
                    compressed.data,
                    compressed.indices,
                    compressed.indptr,
                    totals_of_candidate_cells,
                    medians_of_candidate_genes,
                )

            if noisy_genes_mask is not None:
                compressed[:, noisy_genes_mask] = 0.0

            compressed.data[(compressed.data < min_gene_fold_factor) & (compressed.data > -min_gene_fold_factor)] = 0.0

            ut.eliminate_zeros(compressed)

        else:
            assert dense is not None

            extension_name = f"fold_factor_dense_{dense.dtype}_t"
            extension = getattr(xt, extension_name)

            with ut.timed_step("extensions.fold_factor_dense"):
                extension(
                    dense,
                    totals_of_candidate_cells,
                    medians_of_candidate_genes,
                )

            if noisy_genes_mask is not None:
                dense[:, noisy_genes_mask] = 0.0

            dense[(dense < min_gene_fold_factor) & (dense > -min_gene_fold_factor)] = 0.0

            compressed = sp.csr_matrix(dense)
            assert compressed.has_sorted_indices
            assert compressed.has_canonical_format

        if compressed.nnz == 0:
            acceptable_cells_mask[candidate_cell_indices] = True
        list_of_fold_factors.append(compressed)

    if remaining_cells_count > 0:
        assert remaining_cells_count == np.sum(candidate_of_cells < 0)
        list_of_cell_index_of_rows.append(np.where(candidate_of_cells < 0)[0])
        compressed = sp.csr_matrix(
            ([], [], [0] * (remaining_cells_count + 1)), shape=(remaining_cells_count, genes_count)
        )
        assert compressed.has_sorted_indices
        assert compressed.has_canonical_format
        list_of_fold_factors.append(compressed)

    return list_of_fold_factors, list_of_cell_index_of_rows


@ut.timed_call()
def _construct_fold_factors(
    cells_count: int,
    list_of_fold_factors: List[ut.CompressedMatrix],
    list_of_cell_index_of_rows: List[ut.NumpyVector],
) -> ut.CompressedMatrix:
    cell_index_of_rows = np.concatenate(list_of_cell_index_of_rows)
    if cell_index_of_rows.size == 0:
        return sp.csr_matrix((0, 0))

    cell_row_of_indices = np.empty_like(cell_index_of_rows)
    cell_row_of_indices[cell_index_of_rows] = np.arange(cells_count)

    fold_factors = sp.vstack(list_of_fold_factors, format="csr")
    fold_factors = fold_factors[cell_row_of_indices, :]
    if fold_factors.nnz > 0:
        fold_factors = ut.to_layout(fold_factors, "column_major")
    return fold_factors


@ut.timed_call()
def _filter_genes(
    *,
    cells_count: int,
    genes_count: int,
    fold_factors: ut.CompressedMatrix,
    min_gene_fold_factor: float,
    max_gene_fraction: Optional[float] = None,
) -> ut.NumpyVector:
    ut.timed_parameters(cells=cells_count, genes=genes_count, fold_factors=fold_factors.nnz)
    max_fold_factors_of_genes = ut.max_per(fold_factors, per="column")
    assert max_fold_factors_of_genes.size == genes_count

    mask_of_deviant_genes = max_fold_factors_of_genes >= min_gene_fold_factor
    deviant_gene_fraction = np.sum(mask_of_deviant_genes) / genes_count

    if max_gene_fraction is not None and deviant_gene_fraction > max_gene_fraction:
        if ut.logging_calc():
            ut.log_calc("candidate_deviant_genes", mask_of_deviant_genes)

        quantile_gene_fold_factor = np.quantile(max_fold_factors_of_genes, 1 - max_gene_fraction)
        assert quantile_gene_fold_factor is not None
        ut.log_calc("quantile_gene_fold_factor", quantile_gene_fold_factor)

        if quantile_gene_fold_factor > min_gene_fold_factor:
            min_gene_fold_factor = quantile_gene_fold_factor
            mask_of_deviant_genes = max_fold_factors_of_genes >= min_gene_fold_factor

            fold_factors.data[fold_factors.data < min_gene_fold_factor] = 0
            ut.eliminate_zeros(fold_factors)

    if ut.logging_calc():
        ut.log_calc("deviant_genes", mask_of_deviant_genes)

    deviant_gene_indices = np.where(mask_of_deviant_genes)[0]
    return deviant_gene_indices


@ut.timed_call()
def _fold_ranks(
    *,
    cells_count: int,
    fold_factors: ut.CompressedMatrix,
    deviant_gene_indices: ut.NumpyVector,
) -> ut.NumpyMatrix:
    assert fold_factors.getformat() == "csc"

    deviant_genes_count = deviant_gene_indices.size

    ut.timed_parameters(cells=cells_count, deviant_genes=deviant_genes_count)

    deviant_genes_fold_ranks = np.full((cells_count, deviant_genes_count), cells_count, order="F")
    assert ut.is_layout(deviant_genes_fold_ranks, "column_major")

    for deviant_gene_index, gene_index in enumerate(deviant_gene_indices):
        gene_start_offset = fold_factors.indptr[gene_index]
        gene_stop_offset = fold_factors.indptr[gene_index + 1]

        gene_fold_factors = fold_factors.data[gene_start_offset:gene_stop_offset]
        gene_suspect_cell_indices = fold_factors.indices[gene_start_offset:gene_stop_offset]

        gene_fold_ranks = st.rankdata(gene_fold_factors, method="min")
        gene_fold_ranks *= -1
        gene_fold_ranks += gene_fold_ranks.size + 1

        deviant_genes_fold_ranks[gene_suspect_cell_indices, deviant_gene_index] = gene_fold_ranks

    return deviant_genes_fold_ranks


@ut.timed_call()
def _filter_cells(
    *,
    cells_count: int,
    deviant_genes_fold_ranks: ut.NumpyMatrix,
    deviant_gene_indices: ut.NumpyVector,
    votes_of_deviant_genes: ut.NumpyVector,
    max_deviant_cells_count: float,
) -> Tuple[ut.NumpyVector, bool]:
    min_fold_ranks_of_cells = np.min(deviant_genes_fold_ranks, axis=1)
    assert min_fold_ranks_of_cells.size == cells_count

    threshold_cells_fold_rank = cells_count

    mask_of_deviant_cells = min_fold_ranks_of_cells < threshold_cells_fold_rank
    deviant_cells_count = sum(mask_of_deviant_cells)

    ut.log_calc("deviant_cells", mask_of_deviant_cells)

    did_reach_max_deviant_cells = deviant_cells_count >= max_deviant_cells_count
    if did_reach_max_deviant_cells:
        quantile_cells_fold_rank = np.quantile(min_fold_ranks_of_cells, max_deviant_cells_count / cells_count)
        assert quantile_cells_fold_rank is not None

        ut.log_calc("quantile_cells_fold_rank", quantile_cells_fold_rank)

        if quantile_cells_fold_rank < threshold_cells_fold_rank:
            threshold_cells_fold_rank = quantile_cells_fold_rank

    threshold_cells_fold_rank = max(threshold_cells_fold_rank, 2)

    ut.log_calc("threshold_cells_fold_rank", threshold_cells_fold_rank)
    deviant_votes = deviant_genes_fold_ranks < threshold_cells_fold_rank

    votes_of_deviant_cells = ut.sum_per(ut.to_layout(deviant_votes, "row_major"), per="row")
    assert votes_of_deviant_cells.size == cells_count

    used_votes_of_deviant_genes = ut.sum_per(deviant_votes, per="column")
    assert used_votes_of_deviant_genes.size == deviant_gene_indices.size

    votes_of_deviant_genes[deviant_gene_indices] += used_votes_of_deviant_genes

    return votes_of_deviant_cells, did_reach_max_deviant_cells



"""
Selection
---------
"""

from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut

__all__ = [
    "extract_selected_data",
]


# pylint: disable=dangerous-default-value


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def extract_selected_data(  # pylint: disable=too-many-branches, too-many-statements
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    name: Optional[str] = ".select",
    downsample_min_samples: float = pr.select_downsample_min_samples,
    downsample_min_cell_quantile: float = pr.select_downsample_min_cell_quantile,
    downsample_max_cell_quantile: float = pr.select_downsample_max_cell_quantile,
    min_gene_relative_variance: Optional[float],  # = pr.select_min_gene_relative_variance,
    min_genes: int = pr.select_min_genes,
    min_gene_total: Optional[int] = pr.select_min_gene_total,
    min_gene_top3: Optional[int] = pr.select_min_gene_top3,
    additional_gene_masks: List[str] = ["&~lateral_gene"],
    random_seed: int,
    top_level: bool = True,
) -> AnnData:
    """
    Select a subset of ``what`` (default: {what} data, to compute metacells by.

    When computing metacells (or clustering cells in general), it makes sense to use a subset of the genes for computing
    cell-cell similarity, for both technical (e.g., too low an expression level) and biological (e.g., ignoring
    bookkeeping and cell cycle genes) reasons. The steps provided here are expected to be generically useful, but as
    always specific data sets may require custom gene selection steps on a case-by-case basis.

    **Input**

    A presumably "clean" Annotated ``adata``, where the observations are cells and the variables are
    genes, where ``what`` is a per-variable-per-observation matrix or the name of a
    per-variable-per-observation annotation containing such a matrix.

    Will obey the following annotations in the full ``adata``, if they exist:

    Variable (Gene) Annotations

        ``select_gene``
            If exists, force a mask of genes to use as "select" genes, ignoring everything else.

        ``lateral_gene``
            A boolean mask of genes which are lateral from being chosen as "select" genes based
            on their name.

    **Returns**

    Returns annotated sliced data containing the "select" subset of the original data. By default,
    the ``name`` of this data is {name}. If no selects were selected, return ``None``.

    Also sets the following annotations in the full ``adata``:

    Unstructured Annotations
        ``downsample_samples``
            The target total number of samples in each downsampled cell.

    Observation-Variable (Cell-Gene) Annotations:
        ``downsampled``
            The downsampled data where the total number of samples in each cell is at most ``downsample_samples``.

    Variable (Gene) Annotations
        ``high_total_gene``
            A boolean mask of genes with "high" expression level (unless a ``select_gene`` mask exists).

        ``high_relative_variance_gene``
            A boolean mask of genes with "high" normalized variance, relative to other genes with a similar expression
            level (unless a ``select_gene`` mask exists).

        ``selected_gene``
            A boolean mask of the actually selected genes.

    **Computation Parameters**

    0. If a ``select_gene`` mask exists, just use these genes and go directly to the last step 6.

    1. Invoke :py:func:`metacells.tools.downsample.downsample_cells` to downsample the cells to the same total number of
       UMIs, using the ``downsample_min_samples`` (default: {downsample_min_samples}), ``downsample_min_cell_quantile``
       (default: {downsample_min_cell_quantile}), ``downsample_max_cell_quantile`` (default:
       {downsample_max_cell_quantile}) and the ``random_seed`` (non-zero for reproducible results).

    2. Invoke :py:func:`metacells.tools.high.find_high_total_genes` to select high-expression genes (based on the
       downsampled data), using ``min_gene_total``.

    3. Invoke :py:func:`metacells.tools.high.find_high_relative_variance_genes` to select high-variance genes (based on
       the downsampled data), using ``min_gene_relative_variance``.

    4. Compute the set of genes that pass the above test, as well as match the ``additional_gene_masks`` (default:
       {additional_gene_masks}).

    5. If we found less than ``min_genes`` genes (default: {min_genes}, and ``min_gene_relative_variance`` was
       specified, try to achieve the required minimal number of genes by reducing the ``min_gene_relative_variance``. In
       extreme cases, give up on the relative variance requirement altogether.

    6. Invoke :py:func:`metacells.tools.filter.filter_data` to slice just the selected genes.
    """

    assert min_genes > 0

    tl.downsample_cells(
        adata,
        what,
        downsample_min_samples=downsample_min_samples,
        downsample_min_cell_quantile=downsample_min_cell_quantile,
        downsample_max_cell_quantile=downsample_max_cell_quantile,
        random_seed=random_seed,
    )

    results: Optional[Tuple[AnnData, ut.PandasSeries, ut.PandasSeries]] = None
    if ut.has_data(adata, "select_gene"):
        results = tl.filter_data(
            adata,
            name=name,
            top_level=top_level,
            track_var="full_gene_index",
            var_masks=["&select_gene"],
            mask_var="selected_gene",
        )
        assert results is not None

    else:
        var_masks = []

        if min_gene_top3 is not None:
            var_masks.append("&high_top3_gene")
            tl.find_high_topN_genes(adata, "downsampled", topN=3, min_gene_topN=min_gene_top3)

        if min_gene_total is not None:
            var_masks.append("&high_total_gene")
            tl.find_high_total_genes(adata, "downsampled", min_gene_total=min_gene_total)

        if min_gene_relative_variance is not None:
            var_masks.append("&high_relative_variance_gene")
            tl.find_high_relative_variance_genes(
                adata, "downsampled", min_gene_relative_variance=min_gene_relative_variance
            )

        candidate_genes_mask = tl.combine_masks(adata, var_masks + additional_gene_masks)
        assert candidate_genes_mask is not None
        if np.sum(candidate_genes_mask.values) >= min_genes:
            results = tl.filter_data(
                adata,
                name=name,
                top_level=top_level,
                track_var="full_gene_index",
                var_masks=var_masks + additional_gene_masks,
                mask_var="selected_gene",
            )

    if results is None and min_gene_relative_variance is not None:
        valid_candidate_genes_mask = tl.combine_masks(adata, var_masks[:-1] + additional_gene_masks)
        assert valid_candidate_genes_mask is not None

        is_valid_set = False
        if np.sum(valid_candidate_genes_mask.values) <= min_genes:
            var_masks.pop()

        else:
            while True:
                high_gene_relative_variance = min_gene_relative_variance
                min_gene_relative_variance -= 1 / 8
                tl.find_high_relative_variance_genes(
                    adata, "downsampled", min_gene_relative_variance=min_gene_relative_variance
                )
                valid_candidate_genes_mask = tl.combine_masks(adata, var_masks + additional_gene_masks)
                assert valid_candidate_genes_mask is not None
                if np.sum(valid_candidate_genes_mask.values) >= min_genes:
                    is_valid_set = True
                    break

            for _ in range(4):
                mid_gene_relative_variance = (high_gene_relative_variance + min_gene_relative_variance) / 2.0
                tl.find_high_relative_variance_genes(
                    adata, "downsampled", min_gene_relative_variance=mid_gene_relative_variance
                )
                mid_candidate_genes_mask = tl.combine_masks(adata, var_masks + additional_gene_masks)
                assert mid_candidate_genes_mask is not None
                if np.sum(mid_candidate_genes_mask.values) >= min_genes:
                    min_gene_relative_variance = mid_gene_relative_variance
                    valid_candidate_genes_mask = mid_candidate_genes_mask
                    is_valid_set = True
                else:
                    high_gene_relative_variance = mid_gene_relative_variance
                    is_valid_set = False

        if not is_valid_set:
            ut.set_v_data(adata, "high_relative_variance_gene", valid_candidate_genes_mask.values)

        results = tl.filter_data(
            adata,
            name=name,
            top_level=top_level,
            track_var="full_gene_index",
            var_masks=var_masks + additional_gene_masks,
            mask_var="selected_gene",
        )

    if results is None:
        raise ValueError(f"Failed to select {min_genes} genes")

    return results[0]



"""
Mask
----
"""

from typing import Collection
from typing import Optional

from anndata import AnnData  # type: ignore

import metacells.utilities as ut

__all__ = [
    "combine_masks",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def combine_masks(  # pylint: disable=too-many-branches,too-many-statements
    adata: AnnData,
    masks: Collection[str],
    *,
    invert: bool = False,
    to: Optional[str] = None,
) -> Optional[ut.PandasSeries]:
    """
    Combine different pre-computed masks into a final overall mask.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    If ``to`` (default: {to}) is ``None``, returns the computed mask. Otherwise, sets the
    mask as an annotation (per-variable or per-observation depending on the type of the combined masks).

    **Computation Parameters**

    1. For each of the mask in ``masks``, in order (left to right), fetch it. Silently ignore missing masks if the name
       has a ``?`` suffix. If the first character of the mask name is ``&``, restrict the current mask, otherwise the
       first character must be ``|`` and we'll expand the mask (for the 1st mask, the mask becomes the current mask
       regardless of the 1st character). If the following character is ``~``, first invert the mask before applying it.

    3. If ``invert`` (default: {invert}), invert the final result mask.
    """
    assert len(masks) > 0

    per: Optional[str] = None

    result_mask: Optional[ut.NumpyVector] = None

    for mask_name in masks:
        log_mask_name = mask_name

        if mask_name[0] == "|":
            is_or = True
            mask_name = mask_name[1:]
        elif mask_name[0] == "&":
            is_or = False
            mask_name = mask_name[1:]
        else:
            raise ValueError(f"invalid mask name: {mask_name} (does not start with & or |)")

        if mask_name[0] == "~":
            invert_mask = True
            mask_name = mask_name[1:]
        else:
            invert_mask = False

        if mask_name[-1] == "?":
            must_exist = False
            mask_name = mask_name[:-1]
        else:
            must_exist = True

        if mask_name in adata.obs:
            mask_per = "o"
            mask = ut.get_o_numpy(adata, mask_name, formatter=ut.mask_description) > 0
        elif mask_name in adata.var:
            mask_per = "v"
            mask = ut.get_v_numpy(adata, mask_name, formatter=ut.mask_description) > 0
        else:
            if must_exist:
                raise KeyError(f"unknown mask data: {mask_name}")
            continue

        if mask.dtype != "bool":
            raise ValueError(f"the data: {mask_name} is not a boolean mask")

        if invert_mask:
            mask = ~mask

        if ut.logging_calc():
            ut.log_calc(log_mask_name, mask)

        if per is None:
            per = mask_per
        else:
            if mask_per != per:
                raise ValueError("mixing per-observation and per-variable masks")

        if result_mask is None:
            result_mask = mask
        elif is_or:
            result_mask = result_mask | mask
        else:
            result_mask = result_mask & mask

    if result_mask is None:
        raise ValueError("no masks to combine")

    if invert:
        result_mask = ~result_mask

    if to is None:
        ut.log_return("result", result_mask)
        if per == "o":
            return ut.to_pandas_series(result_mask, index=adata.obs_names)
        assert per == "v"
        return ut.to_pandas_series(result_mask, index=adata.var_names)

    if per == "o":
        ut.set_o_data(adata, to, result_mask)
    else:
        ut.set_v_data(adata, to, result_mask)

    return None
